/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Sat May 18 15:23:21 CEST 2002: included cpu autodetection from Magnus Ekdahl
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

#include "readdb.h"
#include "defaults.h"
#include "others.h"
#include "options.h"
#include "manager.h"
#include "treewalk.h"
#include "shared.h"

int scanmanager(const struct optstruct *opt)
{
	mode_t fmode;
	sigset_t signal_set;
	int ret = 0, method, threads = 0, i, compression = 0;
	struct signature **sigtab;
	char *fullpath = NULL, cwd[200];

    /* w -> whole file */

    if((optc(opt, 'w'))) method = 1;
    else method = 0;

    if(optl(opt, "threads")) {
	threads = atoi(getargl(opt, "threads"));
	if(threads < 0) {
	    mprintf("@Invalid threads number.\n");
	    return 51;
	}
    }
#ifdef C_LINUX
    else /* Autodetecting threads */
      {
	if((threads = detectCpu()) > 1)
	    mprintf("Autodetected %d CPUs. Starting %d threads\n", threads, threads);
	else
	    threads = 0;
      }
#endif

    if(optl(opt, "unzip") || optl(opt, "unrar") || optl(opt, "unace") ||
       optl(opt, "unarj") || optl(opt, "zoo") || optl(opt, "jar") ||
       optl(opt, "lha") || optl(opt, "tar") || optl(opt, "tgz"))
	    compression = 1;


    /* now initialize the database from hardcoded file or --database */

    if(optc(opt, 'd'))
	sigtab = readdb(getargc(opt, 'd'));
    else
	sigtab = readdb(VIRUSDB);

    if(!sigtab) {
	mprintf("@Can't initialize virus database.\n");
	return 50;
    }

    if(compression) {
	/* block all signals */

	sigfillset(&signal_set); /* all signals */
	sigprocmask(SIG_SETMASK, &signal_set, NULL); /* set lock */
    }

    /* check filetype */
    if(!strcmp(opt->filename, "-")) { /* read data from stdin */
	ret = checkstdin(sigtab);

    } else if(strlen(opt->filename) == 0) {

	/* we need full path for some reasons (eg. archive handling) */
	if(!getcwd(cwd, 200)) {
	    mprintf("@Can't get absolute pathname of current working directory.\n");
	    ret = 57;
	} else
	    ret = scandirs(cwd, sigtab, method, threads, opt);

    } else if((fmode = (mode_t) info(opt->filename, 2)) == -1) {
	mprintf("@Can't access file %s\n", opt->filename);
	perror(opt->filename);
	ret = 56;
    } else {
	if(compression && (opt->filename[0] != '/')) {
	    /* we need to complete the path */
	    if(!getcwd(cwd, 200)) {
		mprintf("@Can't get absolute pathname of current working directory.\n");
		return 57;
	    } else {
		fullpath = mcalloc(512, sizeof(char));
		snprintf(fullpath, 512, "%s/%s", cwd, opt->filename);
		mprintf("*Full path: %s\n", fullpath);
	    }
	} else 
	    fullpath = (char *) opt->filename;

	switch(fmode & S_IFMT) {
	    case S_IFREG: 
		ret = scanfile(fullpath, sigtab, method, threads, opt);
		break;

	    case S_IFDIR:
		  ret = scandirs(fullpath, sigtab, method, threads, opt);
		  break;
	    default:
		  mprintf("@Not supported file type (%s)\n", opt->filename);
		  ret = 52;
	}
    }

    if(compression)
	free(fullpath);

    /* clear signature structures */

    for(i = 0; i < ALL; i++) {
	    struct signature *handler, *prev;

	handler = sigtab[i]->next;
	while(handler) {
	    prev = handler;
	    handler = handler->next;
	    if(prev->pattern) free(prev->pattern);
	    if(prev->virname) free(prev->virname);
	    free(prev);
	}
	
	free(sigtab[i]);
    }

    free(sigtab);


    if(ret < 50) { /* hopefully no error detected */
	/* overwrite return code */

	if(claminfo.ifiles)
	    ret = 1;
	else
	    ret = 0;
    }

    return ret;
}

int scanfile(const char *filename, struct signature **sigtab, int method, int thread, const struct optstruct *opt)
{

    /* check the extension */

    if((strbcasestr(filename, ".zip") && optl(opt, "unzip")) 
    || (strbcasestr(filename, ".rar") && optl(opt, "unrar"))
    || (strbcasestr(filename, ".ace") && optl(opt, "unace"))
    || (strbcasestr(filename, ".arj") && optl(opt, "unarj"))
    || (strbcasestr(filename, ".zoo") && optl(opt, "zoo"))
    || (strbcasestr(filename, ".jar") && optl(opt, "jar"))
    || (strbcasestr(filename, ".lzh") && optl(opt, "lha"))
    || (strbcasestr(filename, ".tar") && optl(opt, "tar"))
    || ((strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz")) && optl(opt, "tgz")))
	return(scancompressed(filename, sigtab, method, thread, opt));


    if(!method)
	return(checkfile(filename, sigtab, thread));
    else
	return(checkwholefile(filename, sigtab, thread));
}

int scancompressed(const char *filename, struct signature **sigtab, int method, int thread, const struct optstruct *opt)
{
	int ret = 0;
	char *tmpdir, *userprg;
	struct stat statbuf;
	struct passwd *user = NULL;

    stat(filename, &statbuf);

    if(!S_ISREG(statbuf.st_mode)) {
	mprintf("^Suspected archive %s is not a regular file.\n", filename);
	return 0;
    }


#ifndef C_CYGWIN

    if(!getuid()) {

	if((user = getpwnam(UNPUSER)) == NULL) {
	    mprintf("@Can't get information about user "UNPUSER".\n");
	    exit(60); /* this is critical problem, so we just exit here */
	}

	/* I'm too lazy to check is the change of permissions and mode
	 * needed.
	 * We don't check here are that changes successful, because some
	 * files may be on ro devices (such cdrom).
	 */

	chmod(filename, 0400);
	chown(filename, user->pw_uid, user->pw_gid);
    }

#endif

    /* generate temporary directory */
    tmpdir = gentemp(getargl(opt, "tempdir"));
    if(mkdir(tmpdir, 0700)) {
	mprintf("@Can't create temporary directory %s\n", tmpdir);
	return 0;
    }
    if(user)
	chown(tmpdir, user->pw_uid, user->pw_gid);


    /* unpack file  - as unprivileged user */
    if(strbcasestr(filename, ".zip")) {
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };
	if((userprg = getargl(opt, "unzip")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".rar")) { 
	char *args[] = { "unrar", "x", "-p-", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unrar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unrar", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".ace")) { 
	char *args[] = { "unace", "x", "-y", (char *) filename, NULL };
	if((userprg = getargl(opt, "unace")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unace", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".arj")) { 
	char *args[] = { "unarj", "x", (char *) filename, NULL };
	if((userprg = getargl(opt, "unarj")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unarj", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".zoo")) { 
	char *args[] = { "zoo", "x.O", (char *) filename, NULL };
	if((userprg = getargl(opt, "zoo")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("zoo", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".jar")) { 
	char *args[] = { "unzip", "-P", "clam", "-o", (char *) filename, NULL };
	if((userprg = getargl(opt, "jar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("unzip", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".lzh")) { 
	char *args[] = { "lha", "xf", (char *) filename, NULL };
	if((userprg = getargl(opt, "lha")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("lha", args, tmpdir, user, opt);

    } else if(strbcasestr(filename, ".tar")) { 
	char *args[] = { "tar", "-xpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tar")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("tar", args, tmpdir, user, opt);

    } else if((strbcasestr(filename, ".tar.gz") || strbcasestr(filename, ".tgz"))) {
	char *args[] = { "tar", "-zxpvf", (char *) filename, NULL };
	if((userprg = getargl(opt, "tgz")))
	    ret = clamav_unpack(userprg, args, tmpdir, user, opt);
	else
	    ret = clamav_unpack("tar", args, tmpdir, user, opt);
    }

    /* set proper modes on archive */
    chmod(filename, statbuf.st_mode);
    chown(filename, statbuf.st_uid, statbuf.st_gid);

    /* fix permissions of extracted files */
    fixperms(tmpdir);

    if(!ret) /* execute successful */
	/* notice, that we can't use threadscan() so also scandirs() here */
	ret = treewalk(tmpdir, sigtab, method, thread, opt);

    /* remove the directory  - as clamav */
    clamav_rmdirs(tmpdir);

    /* free tmpdir - it's not necessary now */
    free(tmpdir);

    switch(ret) {
	case -1:
	    mprintf("@Can't fork().\n");
	    exit(61); /* this is critical problem, so we just exit here */
	case -2:
	    mprintf("@Can't execute some unpacker. Check paths and permissions on temporary directory.\n");
	    exit(62); /* this is critical problem, so we just exit here */
	case -3:
	    return 0;
	case 0:
	    /* no viruses found in archive, we scan just in case the same
	     * archive */
	    if(!method)
		return(checkfile(filename, sigtab, thread));
	    else
		return(checkwholefile(filename, sigtab, thread));
	case 1:
	    logg("%s: Infected Archive FOUND\n", filename);
	    mprintf("%s: Infected Archive FOUND\n", filename);
	    return 1;
	default:
	    mprintf("@Strange value (%d) returned in scancompressed()\n", ret);
	    return 0;
    }
}

int scandirs(const char *dirname, struct signature **sigtab, int method, int thread, const struct optstruct *opt)
{

#ifdef THREADS
    if(thread)
	return(threadscan(dirname, sigtab, method, thread, opt));
    else
#endif
	return(treewalk(dirname, sigtab, method, thread, opt));
}
